@@ -25,13 +25,13 @@ APETag layout
         Flags (4 bytes)
         Field Name (? ANSI bytes -- requires NULL terminator -- in range of 0x20 (space) to 0x7E (tilde))
         Value ([Value Size] bytes)
-3) Footer - APE_TAG_FOOTER (32 bytes)    
+3) Footer - APE_TAG_FOOTER (32 bytes)
 **************************************************************************************************/
 
 /**************************************************************************************************
 Notes
 
-When saving images, store the filename (no directory -- i.e. Cover.jpg) in UTF-8 followed 
+When saving images, store the filename (no directory -- i.e. Cover.jpg) in UTF-8 followed
 by a null terminator, followed by the image data.
 
 What saving text lists, delimit the values with a NULL terminator.
@@ -48,9 +48,11 @@ The version of the APE tag
 #define APE_TAG_FIELD_TITLE                     L"Title"
 #define APE_TAG_FIELD_ARTIST                    L"Artist"
 #define APE_TAG_FIELD_ALBUM                     L"Album"
+#define APE_TAG_FIELD_ALBUM_ARTIST              L"Album Artist"
 #define APE_TAG_FIELD_COMMENT                   L"Comment"
 #define APE_TAG_FIELD_YEAR                      L"Year"
 #define APE_TAG_FIELD_TRACK                     L"Track"
+#define APE_TAG_FIELD_DISC                      L"Disc"
 #define APE_TAG_FIELD_GENRE                     L"Genre"
 #define APE_TAG_FIELD_COVER_ART_FRONT           L"Cover Art (front)"
 #define APE_TAG_FIELD_NOTES                     L"Notes"
@@ -69,6 +71,8 @@ The version of the APE tag
 #define APE_TAG_FIELD_COMPOSER                  L"Composer"
 #define APE_TAG_FIELD_KEYWORDS                  L"Keywords"
 #define APE_TAG_FIELD_RATING                    L"Rating"
+#define APE_TAG_FIELD_PUBLISHER                 L"Publisher"
+#define APE_TAG_FIELD_BPM                       L"BPM"
 
 /**************************************************************************************************
 Standard APE tag field values
@@ -78,7 +82,8 @@ Standard APE tag field values
 /**************************************************************************************************
 ID3 v1.1 tag
 **************************************************************************************************/
-#define ID3_TAG_BYTES    128
+#define ID3_TAG_BYTES 128 // equal to sizeof(ID3_TAG)
+#pragma pack(push, 1)
 struct ID3_TAG
 {
     char Header[3];             // should equal 'TAG'
@@ -90,6 +95,7 @@ struct ID3_TAG
     unsigned char Track;        // track
     unsigned char Genre;        // genre
 };
+#pragma pack(pop)
 
 /**************************************************************************************************
 Footer (and header) flags
@@ -111,15 +117,17 @@ Tag field flags
 #define TAG_FIELD_FLAG_DATA_TYPE_EXTERNAL_INFO  (2 << 1)
 #define TAG_FIELD_FLAG_DATA_TYPE_RESERVED       (3 << 1)
 
+#pragma pack(push, 1)
+
 /**************************************************************************************************
 The footer at the end of APE tagged files (can also optionally be at the front of the tag)
 **************************************************************************************************/
+#ifdef APE_ENABLE_TAG_FOOTER
 #define APE_TAG_FOOTER_BYTES    32
-
 class APE_TAG_FOOTER
 {
 protected:
-    char m_cID[8];              // should equal 'APETAGEX'    
+    char m_cID[8];              // should equal 'APETAGEX'
     int m_nVersion;             // equals CURRENT_APE_TAG_VERSION
     int m_nSize;                // the complete size of the tag, including this footer (excludes header)
     int m_nFields;              // the number of fields in the tag
@@ -130,35 +138,36 @@ public:
     APE_TAG_FOOTER(int nFields = 0, int nFieldBytes = 0)
     {
         memcpy(m_cID, "APETAGEX", 8);
-        memset(m_cReserved, 0, 8);
+        APE_CLEAR(m_cReserved);
         m_nFields = nFields;
         m_nFlags = APE_TAG_FLAGS_DEFAULT;
         m_nSize = nFieldBytes + APE_TAG_FOOTER_BYTES;
         m_nVersion = CURRENT_APE_TAG_VERSION;
     }
 
-    int GetTotalTagBytes() { return m_nSize + (GetHasHeader() ? APE_TAG_FOOTER_BYTES : 0); }
-    int GetFieldBytes() { return m_nSize - APE_TAG_FOOTER_BYTES; }
-    int GetFieldsOffset() { return GetHasHeader() ? APE_TAG_FOOTER_BYTES : 0; }
-    int GetNumberFields() { return m_nFields; }
-    bool GetHasHeader() { return (m_nFlags & APE_TAG_FLAG_CONTAINS_HEADER) ? true : false; }
-    bool GetIsHeader() { return (m_nFlags & APE_TAG_FLAG_IS_HEADER) ? true : false; }
-    int GetVersion() { return m_nVersion; }
-
-    bool GetIsValid(bool bAllowHeader)
+    __forceinline int GetTotalTagBytes() const { return m_nSize + (GetHasHeader() ? APE_TAG_FOOTER_BYTES : 0); }
+    __forceinline int GetFieldBytes() const { return m_nSize - APE_TAG_FOOTER_BYTES; }
+    __forceinline int GetFieldsOffset() const { return GetHasHeader() ? APE_TAG_FOOTER_BYTES : 0; }
+    __forceinline int GetNumberFields() const { return m_nFields; }
+    __forceinline bool GetHasHeader() const { return (m_nFlags & APE_TAG_FLAG_CONTAINS_HEADER) ? true : false; }
+    __forceinline bool GetIsHeader() const { return (m_nFlags & APE_TAG_FLAG_IS_HEADER) ? true : false; }
+    __forceinline int GetVersion() const { return m_nVersion; }
+    __forceinline void Empty() { APE_CLEAR(m_cID); }
+    __forceinline bool GetIsValid(bool bAllowHeader) const
     {
-        bool bValid = (strncmp(m_cID, "APETAGEX", 8) == 0) && 
+        bool bValid = (strncmp(m_cID, "APETAGEX", 8) == 0) &&
             (m_nVersion <= CURRENT_APE_TAG_VERSION) &&
             (m_nFields <= 65536) &&
             (m_nSize >= APE_TAG_FOOTER_BYTES) &&
-            (GetFieldBytes() <= (1024 * 1024 * 16));
-        
+            (GetFieldBytes() <= (APE_BYTES_IN_MEGABYTE * 256));
+
         if (bValid && !bAllowHeader && GetIsHeader())
             bValid = false;
 
         return bValid ? true : false;
     }
 };
+#endif
 
 /**************************************************************************************************
 CAPETagField class (an APE tag is an array of these)
@@ -168,36 +177,36 @@ class CAPETagField
 public:
     // create a tag field (use nFieldBytes = -1 for null-terminated strings)
     CAPETagField(const str_utfn * pFieldName, const void * pFieldValue, int nFieldBytes = -1, int nFlags = 0);
-    
-    // destructor
-    ~CAPETagField();
+
+    // destruction
+    virtual ~CAPETagField();
 
     // gets the size of the entire field in bytes (name, value, and metadata)
-    int GetFieldSize();
-    
+    int GetFieldSize() const;
+
     // get the name of the field
-    const str_utfn * GetFieldName();
+    const str_utfn * GetFieldName() const;
 
     // get the value of the field
     const char * GetFieldValue();
-    
+
     // get the size of the value (in bytes)
-    int GetFieldValueSize();
+    int GetFieldValueSize() const;
 
     // get any special flags
-    int GetFieldFlags();
-    
+    int GetFieldFlags() const;
+
     // output the entire field to a buffer (GetFieldSize() bytes)
     int SaveField(char * pBuffer, int nBytes);
 
     // checks to see if the field is read-only
-    bool GetIsReadOnly() { return (m_nFieldFlags & TAG_FIELD_FLAG_READ_ONLY) ? true : false; }
-    bool GetIsUTF8Text() { return ((m_nFieldFlags & TAG_FIELD_FLAG_DATA_TYPE_MASK) == TAG_FIELD_FLAG_DATA_TYPE_TEXT_UTF8) ? true : false; }
+    bool GetIsReadOnly() const;
+    bool GetIsUTF8Text() const;
 
     // set helpers (use with EXTREME caution)
-    void SetFieldFlags(int nFlags) { m_nFieldFlags = nFlags; }
+    void SetFieldFlags(int nFlags);
 
-private:  
+private:
     // helpers
     void Save32(char* pBuffer, int nValue);
 
@@ -209,35 +218,93 @@ private:
 };
 
 /**************************************************************************************************
-CAPETag class
+IAPETag interface
 **************************************************************************************************/
-class CAPETag
+class IAPETag
 {
 public:
-    // create an APE tag 
+
+    virtual ~IAPETag() { }
+
+    // save the tag to the I/O source (bUseOldID3 forces it to save as an ID3v1.1 tag instead of an APE tag)
+    virtual int Save(bool bUseOldID3 = false) = 0;
+
+    // removes any tags from the file (bUpdate determines whether is should re-analyze after removing the tag)
+    virtual int Remove(bool bUpdate = true) = 0;
+
+    // sets the value of a field (use nFieldBytes = -1 for null terminated strings)
+    // note: using NULL or "" for a string type will remove the field
+    virtual int SetFieldString(const str_utfn * pFieldName, const str_utfn * pFieldValue, const str_utfn * pListDelimiter = APE_NULL) = 0;
+    virtual int SetFieldString(const str_utfn * pFieldName, const char * pFieldValue, bool bAlreadyUTF8Encoded, const str_utfn * pListDelimiter = APE_NULL) = 0;
+    virtual int SetFieldBinary(const str_utfn * pFieldName, const void * pFieldValue, intn nFieldBytes, int nFieldFlags) = 0;
+
+    // gets the value of a field (returns -1 and an empty buffer if the field doesn't exist)
+    virtual int GetFieldBinary(const str_utfn * pFieldName, void * pBuffer, int * pBufferBytes) = 0;
+    virtual int GetFieldString(const str_utfn * pFieldName, str_utfn * pBuffer, int * pBufferCharacters, const str_utfn * pListDelimiter = L"; ") = 0;
+    virtual int GetFieldString(const str_utfn * pFieldName, str_ansi * pBuffer, int * pBufferCharacters, bool bUTF8Encode = false) = 0;
+
+    // remove a specific field
+    virtual int RemoveField(const str_utfn * pFieldName) = 0;
+    virtual int RemoveField(int nIndex) = 0;
+
+    // clear all the fields
+    virtual int ClearFields() = 0;
+
+    // see if we've been analyzed (we do lazy analysis)
+    virtual bool GetAnalyzed() = 0;
+
+    // get the total tag bytes in the file from the last analyze
+    // need to call Save() then Analyze() to update any changes
+    virtual int GetTagBytes() = 0;
+
+    // fills in an ID3_TAG using the current fields (useful for quickly converting the tag)
+    virtual int CreateID3Tag(ID3_TAG * pID3Tag) = 0;
+
+    // see whether the file has an ID3 or APE tag
+    virtual bool GetHasID3Tag() = 0;
+    virtual bool GetHasAPETag() = 0;
+    virtual int GetAPETagVersion() = 0;
+
+    // gets a desired tag field (returns NULL if not found)
+    // again, be careful, because this a pointer to the actual field in this class
+    virtual CAPETagField* GetTagField(const str_utfn * pFieldName) = 0;
+    virtual CAPETagField* GetTagField(int nIndex) = 0;
+
+    // options
+    virtual void SetIgnoreReadOnly(bool bIgnoreReadOnly) = 0;
+
+};
+
+/**************************************************************************************************
+CAPETag class
+**************************************************************************************************/
+class CAPETag : public IAPETag
+{
+public:
+    // create an APE tag
     // bAnalyze determines whether it will analyze immediately or on the first request
     // be careful with multiple threads / file pointer movement if you don't analyze immediately
-    CAPETag(CIO * pIO, bool bAnalyze = true);
+    CAPETag(CIO * pIO, bool bAnalyze = true, bool bCheckForID3v1 = true);
     CAPETag(const str_utfn * pFilename, bool bAnalyze = true);
-    
+
     // destructor
-    ~CAPETag();
+    virtual ~CAPETag();
 
     // save the tag to the I/O source (bUseOldID3 forces it to save as an ID3v1.1 tag instead of an APE tag)
     int Save(bool bUseOldID3 = false);
-    
+
     // removes any tags from the file (bUpdate determines whether is should re-analyze after removing the tag)
     int Remove(bool bUpdate = true);
 
     // sets the value of a field (use nFieldBytes = -1 for null terminated strings)
     // note: using NULL or "" for a string type will remove the field
-    int SetFieldString(const str_utfn * pFieldName, const str_utfn * pFieldValue, const str_utfn * pListDelimiter = NULL);
-    int SetFieldString(const str_utfn * pFieldName, const char * pFieldValue, bool bAlreadyUTF8Encoded, const str_utfn * pListDelimiter = NULL);
+    int SetFieldString(const str_utfn * pFieldName, const str_utfn * pFieldValue, const str_utfn * pListDelimiter = APE_NULL);
+    int SetFieldString(const str_utfn * pFieldName, const char * pFieldValue, bool bAlreadyUTF8Encoded, const str_utfn * pListDelimiter = APE_NULL);
     int SetFieldBinary(const str_utfn * pFieldName, const void * pFieldValue, intn nFieldBytes, int nFieldFlags);
 
     // gets the value of a field (returns -1 and an empty buffer if the field doesn't exist)
     int GetFieldBinary(const str_utfn * pFieldName, void * pBuffer, int * pBufferBytes);
-    int GetFieldString(const str_utfn * pFieldName, str_utfn * pBuffer, int * pBufferCharacters, const str_utfn * pListDelimiter = _T("; "));
+    int GetFieldString(const str_utfn * pFieldName, str_utfn * pBuffer, int * pBufferCharacters, const str_utfn * pListDelimiter = L"; ");
     int GetFieldString(const str_utfn * pFieldName, str_ansi * pBuffer, int * pBufferCharacters, bool bUTF8Encode = false);
 
     // remove a specific field
@@ -246,9 +313,9 @@ public:
 
     // clear all the fields
     int ClearFields();
-    
+
     // see if we've been analyzed (we do lazy analysis)
-    bool GetAnalyzed() { return m_bAnalyzed; }
+    bool GetAnalyzed();
 
     // get the total tag bytes in the file from the last analyze
     // need to call Save() then Analyze() to update any changes
@@ -258,9 +325,9 @@ public:
     int CreateID3Tag(ID3_TAG * pID3Tag);
 
     // see whether the file has an ID3 or APE tag
-    bool GetHasID3Tag() { if (!m_bAnalyzed) { Analyze(); } return m_bHasID3Tag; }
-    bool GetHasAPETag() { if (!m_bAnalyzed) { Analyze(); } return m_bHasAPETag; }
-    int GetAPETagVersion() { return GetHasAPETag() ? m_nAPETagVersion : -1; }
+    bool GetHasID3Tag();
+    bool GetHasAPETag();
+    int GetAPETagVersion();
 
     // gets a desired tag field (returns NULL if not found)
     // again, be careful, because this a pointer to the actual field in this class
@@ -268,7 +335,7 @@ public:
     CAPETagField * GetTagField(int nIndex);
 
     // options
-    void SetIgnoreReadOnly(bool bIgnoreReadOnly) { m_bIgnoreReadOnly = bIgnoreReadOnly; }
+    void SetIgnoreReadOnly(bool bIgnoreReadOnly);
 
     // statics
     static const int s_nID3GenreUndefined = 255;
@@ -290,15 +357,18 @@ private:
 
     // private data
     CSmartPtr<CIO> m_spIO;
-    bool m_bAnalyzed;
     int m_nTagBytes;
     int m_nFields;
     int m_nAllocatedFields;
+    int m_nAPETagVersion;
     CAPETagField ** m_aryFields;
     bool m_bHasAPETag;
-    int m_nAPETagVersion;
+    bool m_bAnalyzed;
     bool m_bHasID3Tag;
     bool m_bIgnoreReadOnly;
+    bool m_bCheckForID3v1;
 };
 
-}
\ No newline at end of file
+#pragma pack(pop)
+
+}
